@@ -1,3 +1,5 @@
+// SPDX-License-Identifier: Apache-2.0 OR MIT
+
 use crate::{uapi, Access, CompatError};
 
 #[cfg(test)]
@@ -64,6 +66,9 @@ pub enum ABI {
     /// Fifth Landlock ABI, introduced with
     /// [Linux 6.10](https://git.kernel.org/stable/c/2fc0e7892c10734c1b7c613ef04836d57d4676d5).
     V5 = 5,
+    /// Sixth Landlock ABI, introduced with
+    /// [Linux 6.12](https://git.kernel.org/stable/c/e1b061b444fb01c237838f0d8238653afe6a8094).
+    V6 = 6,
 }
 
 impl ABI {
@@ -80,8 +85,16 @@ impl ABI {
         })
     }
 
-    // There is no way to not publicly expose an implementation of an external trait such as
-    // From<i32>.  See RFC https://github.com/rust-lang/rfcs/pull/2529
+    #[cfg(test)]
+    fn is_known(value: i32) -> bool {
+        value > 0 && value < ABI::COUNT as i32
+    }
+}
+
+/// Converting from an integer to an ABI should only be used for testing.
+/// Indeed, manually setting the ABI can lead to inconsistent and unexpected behaviors.
+/// Instead, just use the appropriate access rights, this library will handle the rest.
+impl From<i32> for ABI {
     fn from(value: i32) -> ABI {
         match value {
             // The only possible error values should be EOPNOTSUPP and ENOSYS, but let's interpret
@@ -91,15 +104,11 @@ impl ABI {
             2 => ABI::V2,
             3 => ABI::V3,
             4 => ABI::V4,
+            5 => ABI::V5,
             // Returns the greatest known ABI.
-            _ => ABI::V5,
+            _ => ABI::V6,
         }
     }
-
-    #[cfg(test)]
-    fn is_known(value: i32) -> bool {
-        value > 0 && value < ABI::COUNT as i32
-    }
 }
 
 #[test]
@@ -146,7 +155,7 @@ lazy_static! {
                 panic!("Unknown ABI: {n}");
             }
         }
-        Err(std::env::VarError::NotPresent) => ABI::iter().last().unwrap(),
+        Err(std::env::VarError::NotPresent) => ABI::new_current(),
         Err(e) => panic!("Failed to read LANDLOCK_CRATE_TEST_ABI: {e}"),
     };
 }
@@ -166,19 +175,15 @@ pub(crate) fn can_emulate(mock: ABI, partial_support: ABI, full_support: Option<
 pub(crate) fn get_errno_from_landlock_status() -> Option<i32> {
     use std::io::Error;
 
-    if unsafe {
-        uapi::landlock_create_ruleset(std::ptr::null(), 0, uapi::LANDLOCK_CREATE_RULESET_VERSION)
-    } < 0
-    {
-        match Error::last_os_error().raw_os_error() {
+    match ABI::new_current() {
+        ABI::Unsupported => match Error::last_os_error().raw_os_error() {
             // Returns ENOSYS when the kernel is not built with Landlock support,
             // or EOPNOTSUPP when Landlock is supported but disabled at boot time.
             ret @ Some(libc::ENOSYS | libc::EOPNOTSUPP) => ret,
             // Other values can only come from bogus seccomp filters or debug tampering.
             _ => unreachable!(),
-        }
-    } else {
-        None
+        },
+        _ => None,
     }
 }
 
@@ -189,7 +194,16 @@ fn current_kernel_abi() {
     // to the Landlock ABI version supported by your kernel.  With a missing variable, the latest
     // Landlock ABI version known by this crate is automatically set.
     // From Linux 5.13 to 5.18, you need to run: LANDLOCK_CRATE_TEST_ABI=1 cargo test
-    assert_eq!(*TEST_ABI, ABI::new_current());
+    let test_abi = *TEST_ABI;
+    let current_abi = ABI::new_current();
+    println!(
+        "Current kernel version: {}",
+        std::fs::read_to_string("/proc/version")
+            .unwrap_or_else(|_| "unknown".into())
+            .trim()
+    );
+    println!("Expected Landlock ABI {test_abi:?} whereas the current ABI is {current_abi:#?}");
+    assert_eq!(test_abi, current_abi);
 }
 
 // CompatState is not public outside this crate.
@@ -389,7 +403,7 @@ pub trait Compatible: Sized + private::OptionCompatLevelMut {
     ///         // However, this ruleset may also handle other (future) access rights
     ///         // if they are supported by the running kernel.
     ///         .set_compatibility(CompatLevel::BestEffort)
-    ///         .handle_access(AccessFs::from_all(ABI::V5))?
+    ///         .handle_access(AccessFs::from_all(ABI::V6))?
     ///         .create()?)
     /// }
     /// ```
@@ -418,7 +432,7 @@ pub trait Compatible: Sized + private::OptionCompatLevelMut {
     ///         // if they are supported by the running kernel,
     ///         // but without returning any error otherwise.
     ///         .set_compatibility(CompatLevel::BestEffort)
-    ///         .handle_access(AccessFs::from_all(ABI::V5))?
+    ///         .handle_access(AccessFs::from_all(ABI::V6))?
     ///         .create()?)
     /// }
     /// ```
