@@ -1,9 +1,11 @@
+// SPDX-License-Identifier: Apache-2.0 OR MIT
+
 use crate::compat::private::OptionCompatLevelMut;
 use crate::{
     uapi, Access, AddRuleError, AddRulesError, CompatError, CompatLevel, CompatResult, CompatState,
-    Compatible, HandleAccessError, HandleAccessesError, PathBeneathError, PathFdError,
-    PrivateAccess, PrivateRule, Rule, Ruleset, RulesetCreated, RulesetError, TailoredCompatLevel,
-    TryCompat, ABI,
+    Compatible, HandleAccessError, HandleAccessesError, HandledAccess, PathBeneathError,
+    PathFdError, PrivateHandledAccess, PrivateRule, Rule, Ruleset, RulesetCreated, RulesetError,
+    TailoredCompatLevel, TryCompat, ABI,
 };
 use enumflags2::{bitflags, make_bitflags, BitFlags};
 use std::fs::OpenOptions;
@@ -57,6 +59,10 @@ pub enum AccessFs {
     /// Execute a file.
     Execute = uapi::LANDLOCK_ACCESS_FS_EXECUTE as u64,
     /// Open a file with write access.
+    ///
+    /// # Note
+    ///
+    /// Certain operations (such as [`std::fs::write`]) may also require [`AccessFs::Truncate`] since [`ABI::V3`].
     WriteFile = uapi::LANDLOCK_ACCESS_FS_WRITE_FILE as u64,
     /// Open a file with read access.
     ReadFile = uapi::LANDLOCK_ACCESS_FS_READ_FILE as u64,
@@ -106,7 +112,7 @@ impl AccessFs {
     pub fn from_read(abi: ABI) -> BitFlags<Self> {
         match abi {
             ABI::Unsupported => BitFlags::EMPTY,
-            ABI::V1 | ABI::V2 | ABI::V3 | ABI::V4 | ABI::V5 => make_bitflags!(AccessFs::{
+            ABI::V1 | ABI::V2 | ABI::V3 | ABI::V4 | ABI::V5 | ABI::V6 => make_bitflags!(AccessFs::{
                 Execute
                 | ReadFile
                 | ReadDir
@@ -134,7 +140,7 @@ impl AccessFs {
             }),
             ABI::V2 => Self::from_write(ABI::V1) | AccessFs::Refer,
             ABI::V3 | ABI::V4 => Self::from_write(ABI::V2) | AccessFs::Truncate,
-            ABI::V5 => Self::from_write(ABI::V4) | AccessFs::IoctlDev,
+            ABI::V5 | ABI::V6 => Self::from_write(ABI::V4) | AccessFs::IoctlDev,
         }
     }
 
@@ -155,7 +161,9 @@ fn consistent_access_fs_rw() {
     }
 }
 
-impl PrivateAccess for AccessFs {
+impl HandledAccess for AccessFs {}
+
+impl PrivateHandledAccess for AccessFs {
     fn ruleset_handle_access(
         ruleset: &mut Ruleset,
         access: BitFlags<Self>,
@@ -310,7 +318,7 @@ fn path_beneath_try_compat_children() {
     // Do not actually perform any syscall.
     ruleset.compat.state = CompatState::Dummy;
     assert!(matches!(
-        RulesetCreated::new(ruleset, -1)
+        RulesetCreated::new(ruleset, None)
             .set_compatibility(CompatLevel::HardRequirement)
             .add_rule(PathBeneath::new(PathFd::new("/dev/null").unwrap(), access_file))
             .unwrap_err(),
@@ -324,7 +332,7 @@ fn path_beneath_try_compat_children() {
     // Do not actually perform any syscall.
     ruleset.compat.state = CompatState::Dummy;
     assert!(matches!(
-        RulesetCreated::new(ruleset, -1)
+        RulesetCreated::new(ruleset, None)
             .set_compatibility(CompatLevel::HardRequirement)
             .add_rule(PathBeneath::new(PathFd::new("/dev/null").unwrap(), access_file))
             .unwrap_err(),
@@ -540,8 +548,11 @@ fn path_fd() {
 
 /// Helper to quickly create an iterator of PathBeneath rules.
 ///
-/// Silently ignores paths that cannot be opened, and automatically adjust access rights according
-/// to file types when possible.
+/// # Note
+///
+/// Landlock rules operate on file descriptors, not paths. This is a helper to create rules based on paths. **It
+/// silently ignores paths that cannot be opened**, and automatically adjusts access rights according to file types when
+/// possible.
 ///
 /// # Example
 ///
