@@ -1,13 +1,14 @@
+// SPDX-License-Identifier: Apache-2.0 OR MIT
+
 use crate::compat::private::OptionCompatLevelMut;
 use crate::{
-    uapi, Access, AccessFs, AccessNet, AddRuleError, AddRulesError, BitFlags, CompatLevel,
-    CompatState, Compatibility, Compatible, CreateRulesetError, RestrictSelfError, RulesetError,
-    TryCompat,
+    uapi, AccessFs, AccessNet, AddRuleError, AddRulesError, BitFlags, CompatLevel, CompatState,
+    Compatibility, Compatible, CreateRulesetError, HandledAccess, PrivateHandledAccess,
+    RestrictSelfError, RulesetError, Scope, ScopeError, TryCompat,
 };
-use libc::close;
 use std::io::Error;
 use std::mem::size_of_val;
-use std::os::unix::io::RawFd;
+use std::os::unix::io::{AsRawFd, FromRawFd, OwnedFd};
 
 #[cfg(test)]
 use crate::*;
@@ -15,7 +16,7 @@ use crate::*;
 // Public interface without methods and which is impossible to implement outside this crate.
 pub trait Rule<T>: PrivateRule<T>
 where
-    T: Access,
+    T: HandledAccess,
 {
 }
 
@@ -23,7 +24,7 @@ where
 pub trait PrivateRule<T>
 where
     Self: TryCompat<T> + Compatible,
-    T: Access,
+    T: HandledAccess,
 {
     const TYPE_ID: uapi::landlock_rule_type;
 
@@ -173,8 +174,10 @@ fn support_no_new_privs() -> bool {
 pub struct Ruleset {
     pub(crate) requested_handled_fs: BitFlags<AccessFs>,
     pub(crate) requested_handled_net: BitFlags<AccessNet>,
+    pub(crate) requested_scoped: BitFlags<Scope>,
     pub(crate) actual_handled_fs: BitFlags<AccessFs>,
     pub(crate) actual_handled_net: BitFlags<AccessNet>,
+    pub(crate) actual_scoped: BitFlags<Scope>,
     pub(crate) compat: Compatibility,
 }
 
@@ -184,8 +187,10 @@ impl From<Compatibility> for Ruleset {
             // Non-working default handled FS accesses to force users to set them explicitely.
             requested_handled_fs: Default::default(),
             requested_handled_net: Default::default(),
+            requested_scoped: Default::default(),
             actual_handled_fs: Default::default(),
             actual_handled_net: Default::default(),
+            actual_scoped: Default::default(),
             compat,
         }
     }
@@ -256,6 +261,7 @@ impl Ruleset {
                     assert!(
                         !self.requested_handled_fs.is_empty()
                             || !self.requested_handled_net.is_empty()
+                            || !self.requested_scoped.is_empty()
                     );
 
                     // CompatState::No should be handled as CompatState::Dummy because it is not
@@ -265,22 +271,28 @@ impl Ruleset {
                         CompatLevel::HardRequirement => {
                             Err(CreateRulesetError::MissingHandledAccess)
                         }
-                        _ => Ok(RulesetCreated::new(self, -1)),
+                        _ => Ok(RulesetCreated::new(self, None)),
                     }
                 }
                 CompatState::Full | CompatState::Partial => {
                     // There is at least one actual handled access.
                     #[cfg(test)]
                     assert!(
-                        !self.actual_handled_fs.is_empty() || !self.actual_handled_net.is_empty()
+                        !self.actual_handled_fs.is_empty()
+                            || !self.actual_handled_net.is_empty()
+                            || !self.actual_scoped.is_empty()
                     );
 
                     let attr = uapi::landlock_ruleset_attr {
                         handled_access_fs: self.actual_handled_fs.bits(),
                         handled_access_net: self.actual_handled_net.bits(),
+                        scoped: self.actual_scoped.bits(),
                     };
                     match unsafe { uapi::landlock_create_ruleset(&attr, size_of_val(&attr), 0) } {
-                        fd if fd >= 0 => Ok(RulesetCreated::new(self, fd)),
+                        fd if fd >= 0 => Ok(RulesetCreated::new(
+                            self,
+                            Some(unsafe { OwnedFd::from_raw_fd(fd) }),
+                        )),
                         _ => Err(CreateRulesetError::CreateRulesetCall {
                             source: Error::last_os_error(),
                         }),
@@ -339,11 +351,37 @@ pub trait RulesetAttr: Sized + AsMut<Ruleset> + Compatible {
     fn handle_access<T, U>(mut self, access: T) -> Result<Self, RulesetError>
     where
         T: Into<BitFlags<U>>,
-        U: Access,
+        U: HandledAccess + PrivateHandledAccess,
     {
         U::ruleset_handle_access(self.as_mut(), access.into())?;
         Ok(self)
     }
+
+    /// Attempts to add a set of scopes that will be supported by this ruleset.
+    /// Consecutive calls to `scope()` will be interpreted as logical ORs
+    /// with the previous scopes.
+    ///
+    /// On error, returns a wrapped [`ScopeError`](crate::ScopeError).
+    /// E.g., `RulesetError::Scope(ScopeError)`
+    fn scope<T>(mut self, scope: T) -> Result<Self, RulesetError>
+    where
+        T: Into<BitFlags<Scope>>,
+    {
+        let scope = scope.into();
+        let ruleset = self.as_mut();
+        ruleset.requested_scoped |= scope;
+        if let Some(a) = scope
+            .try_compat(
+                ruleset.compat.abi(),
+                ruleset.compat.level,
+                &mut ruleset.compat.state,
+            )
+            .map_err(ScopeError::Compat)?
+        {
+            ruleset.actual_scoped |= a;
+        }
+        Ok(self)
+    }
 }
 
 impl RulesetAttr for Ruleset {}
@@ -376,13 +414,21 @@ fn ruleset_attr() {
 
 #[test]
 fn ruleset_created_handle_access_fs() {
+    let access = make_bitflags!(AccessFs::{Execute | ReadDir});
+
     // Tests AccessFs::ruleset_handle_access()
+    let ruleset = Ruleset::from(ABI::V1).handle_access(access).unwrap();
+    assert_eq!(ruleset.requested_handled_fs, access);
+    assert_eq!(ruleset.actual_handled_fs, access);
+
+    // Tests composition (binary OR) of handled accesses.
     let ruleset = Ruleset::from(ABI::V1)
         .handle_access(AccessFs::Execute)
         .unwrap()
         .handle_access(AccessFs::ReadDir)
+        .unwrap()
+        .handle_access(AccessFs::Execute)
         .unwrap();
-    let access = make_bitflags!(AccessFs::{Execute | ReadDir});
     assert_eq!(ruleset.requested_handled_fs, access);
     assert_eq!(ruleset.actual_handled_fs, access);
 
@@ -414,6 +460,17 @@ fn ruleset_created_handle_access_net_tcp() {
     assert_eq!(ruleset.requested_handled_net, access);
     assert_eq!(ruleset.actual_handled_net, access);
 
+    // Tests composition (binary OR) of handled accesses.
+    let ruleset = Ruleset::from(ABI::V4)
+        .handle_access(AccessNet::BindTcp)
+        .unwrap()
+        .handle_access(AccessNet::ConnectTcp)
+        .unwrap()
+        .handle_access(AccessNet::BindTcp)
+        .unwrap();
+    assert_eq!(ruleset.requested_handled_net, access);
+    assert_eq!(ruleset.actual_handled_net, access);
+
     // Tests that only the required handled accesses are reported as incompatible:
     // access should not contains AccessNet::BindTcp.
     assert!(matches!(Ruleset::from(ABI::Unsupported)
@@ -428,6 +485,82 @@ fn ruleset_created_handle_access_net_tcp() {
     ));
 }
 
+#[test]
+fn ruleset_created_scope() {
+    let scopes = make_bitflags!(Scope::{AbstractUnixSocket | Signal});
+
+    // Tests Ruleset::scope() with ABI that doesn't support scopes.
+    let ruleset = Ruleset::from(ABI::V5).scope(scopes).unwrap();
+    assert_eq!(ruleset.requested_scoped, scopes);
+    assert_eq!(ruleset.actual_scoped, BitFlags::<Scope>::EMPTY);
+
+    // Tests Ruleset::scope() with ABI that supports scopes.
+    let ruleset = Ruleset::from(ABI::V6).scope(scopes).unwrap();
+    assert_eq!(ruleset.requested_scoped, scopes);
+    assert_eq!(ruleset.actual_scoped, scopes);
+
+    // Tests composition (binary OR) of scopes.
+    let ruleset = Ruleset::from(ABI::V6)
+        .scope(Scope::AbstractUnixSocket)
+        .unwrap()
+        .scope(Scope::Signal)
+        .unwrap()
+        .scope(Scope::AbstractUnixSocket)
+        .unwrap();
+    assert_eq!(ruleset.requested_scoped, scopes);
+    assert_eq!(ruleset.actual_scoped, scopes);
+
+    // Tests that only the required scopes are reported as incompatible:
+    // scope should not contain Scope::AbstractUnixSocket.
+    assert!(matches!(Ruleset::from(ABI::Unsupported)
+        .scope(Scope::AbstractUnixSocket)
+        .unwrap()
+        .set_compatibility(CompatLevel::HardRequirement)
+        .scope(Scope::Signal)
+        .unwrap_err(),
+        RulesetError::Scope(ScopeError::Compat(
+            CompatError::Access(AccessError::Incompatible { access })
+        )) if access == Scope::Signal
+    ));
+}
+
+#[test]
+fn ruleset_created_fs_net_scope() {
+    let access_fs = make_bitflags!(AccessFs::{Execute | ReadDir});
+    let access_net = make_bitflags!(AccessNet::{BindTcp | ConnectTcp});
+    let scopes = make_bitflags!(Scope::{AbstractUnixSocket | Signal});
+
+    // Tests composition (binary OR) of handled accesses.
+    let ruleset = Ruleset::from(ABI::V5)
+        .handle_access(access_fs)
+        .unwrap()
+        .scope(scopes)
+        .unwrap()
+        .handle_access(access_net)
+        .unwrap();
+    assert_eq!(ruleset.requested_handled_fs, access_fs);
+    assert_eq!(ruleset.actual_handled_fs, access_fs);
+    assert_eq!(ruleset.requested_handled_net, access_net);
+    assert_eq!(ruleset.actual_handled_net, access_net);
+    assert_eq!(ruleset.requested_scoped, scopes);
+    assert_eq!(ruleset.actual_scoped, BitFlags::<Scope>::EMPTY);
+
+    // Tests composition (binary OR) of handled accesses and scopes.
+    let ruleset = Ruleset::from(ABI::V6)
+        .handle_access(access_fs)
+        .unwrap()
+        .scope(scopes)
+        .unwrap()
+        .handle_access(access_net)
+        .unwrap();
+    assert_eq!(ruleset.requested_handled_fs, access_fs);
+    assert_eq!(ruleset.actual_handled_fs, access_fs);
+    assert_eq!(ruleset.requested_handled_net, access_net);
+    assert_eq!(ruleset.actual_handled_net, access_net);
+    assert_eq!(ruleset.requested_scoped, scopes);
+    assert_eq!(ruleset.actual_scoped, scopes);
+}
+
 impl OptionCompatLevelMut for RulesetCreated {
     fn as_option_compat_level_mut(&mut self) -> &mut Option<CompatLevel> {
         &mut self.compat.level
@@ -451,7 +584,7 @@ pub trait RulesetCreatedAttr: Sized + AsMut<RulesetCreated> + Compatible {
     fn add_rule<T, U>(mut self, rule: T) -> Result<Self, RulesetError>
     where
         T: Rule<U>,
-        U: Access,
+        U: HandledAccess + PrivateHandledAccess,
     {
         let body = || -> Result<Self, AddRulesError> {
             let self_ref = self.as_mut();
@@ -469,15 +602,20 @@ pub trait RulesetCreatedAttr: Sized + AsMut<RulesetCreated> + Compatible {
             };
             match self_ref.compat.state {
                 CompatState::Init | CompatState::No | CompatState::Dummy => Ok(self),
-                CompatState::Full | CompatState::Partial => match unsafe {
-                    uapi::landlock_add_rule(self_ref.fd, T::TYPE_ID, compat_rule.as_ptr(), 0)
-                } {
-                    0 => Ok(self),
-                    _ => Err(AddRuleError::<U>::AddRuleCall {
-                        source: Error::last_os_error(),
+                CompatState::Full | CompatState::Partial => {
+                    #[cfg(test)]
+                    assert!(self_ref.fd.is_some());
+                    let fd = self_ref.fd.as_ref().map(|f| f.as_raw_fd()).unwrap_or(-1);
+                    match unsafe {
+                        uapi::landlock_add_rule(fd, T::TYPE_ID, compat_rule.as_ptr(), 0)
+                    } {
+                        0 => Ok(self),
+                        _ => Err(AddRuleError::<U>::AddRuleCall {
+                            source: Error::last_os_error(),
+                        }
+                        .into()),
                     }
-                    .into()),
-                },
+                }
             }
         };
         Ok(body()?)
@@ -561,7 +699,7 @@ pub trait RulesetCreatedAttr: Sized + AsMut<RulesetCreated> + Compatible {
     where
         I: IntoIterator<Item = Result<T, E>>,
         T: Rule<U>,
-        U: Access,
+        U: HandledAccess + PrivateHandledAccess,
         E: From<RulesetError>,
     {
         for rule in rules {
@@ -584,7 +722,7 @@ pub trait RulesetCreatedAttr: Sized + AsMut<RulesetCreated> + Compatible {
 /// Ruleset created with [`Ruleset::create()`].
 #[cfg_attr(test, derive(Debug))]
 pub struct RulesetCreated {
-    fd: RawFd,
+    fd: Option<OwnedFd>,
     no_new_privs: bool,
     pub(crate) requested_handled_fs: BitFlags<AccessFs>,
     pub(crate) requested_handled_net: BitFlags<AccessNet>,
@@ -592,7 +730,7 @@ pub struct RulesetCreated {
 }
 
 impl RulesetCreated {
-    pub(crate) fn new(ruleset: Ruleset, fd: RawFd) -> Self {
+    pub(crate) fn new(ruleset: Ruleset, fd: Option<OwnedFd>) -> Self {
         // The compatibility state is initialized by Ruleset::create().
         #[cfg(test)]
         assert!(!matches!(ruleset.compat.state, CompatState::Init));
@@ -662,7 +800,11 @@ impl RulesetCreated {
                     no_new_privs: enforced_nnp,
                 }),
                 CompatState::Full | CompatState::Partial => {
-                    match unsafe { uapi::landlock_restrict_self(self.fd, 0) } {
+                    #[cfg(test)]
+                    assert!(self.fd.is_some());
+                    // Does not consume ruleset FD, which will be automatically closed after this block.
+                    let fd = self.fd.as_ref().map(|f| f.as_raw_fd()).unwrap_or(-1);
+                    match unsafe { uapi::landlock_restrict_self(fd, 0) } {
                         0 => {
                             self.compat.update(CompatState::Full);
                             Ok(RestrictionStatus {
@@ -687,13 +829,7 @@ impl RulesetCreated {
     /// On error, returns [`std::io::Error`].
     pub fn try_clone(&self) -> std::io::Result<Self> {
         Ok(RulesetCreated {
-            fd: match self.fd {
-                -1 => -1,
-                self_fd => match unsafe { libc::fcntl(self_fd, libc::F_DUPFD_CLOEXEC, 0) } {
-                    dup_fd if dup_fd >= 0 => dup_fd,
-                    _ => return Err(Error::last_os_error()),
-                },
-            },
+            fd: self.fd.as_ref().map(|f| f.try_clone()).transpose()?,
             no_new_privs: self.no_new_privs,
             requested_handled_fs: self.requested_handled_fs,
             requested_handled_net: self.requested_handled_net,
@@ -702,14 +838,23 @@ impl RulesetCreated {
     }
 }
 
-impl Drop for RulesetCreated {
-    fn drop(&mut self) {
-        if self.fd >= 0 {
-            unsafe { close(self.fd) };
-        }
+impl From<RulesetCreated> for Option<OwnedFd> {
+    fn from(ruleset: RulesetCreated) -> Self {
+        ruleset.fd
     }
 }
 
+#[test]
+fn ruleset_created_ownedfd_none() {
+    let ruleset = Ruleset::from(ABI::Unsupported)
+        .handle_access(AccessFs::Execute)
+        .unwrap()
+        .create()
+        .unwrap();
+    let fd: Option<OwnedFd> = ruleset.into();
+    assert!(fd.is_none());
+}
+
 impl AsMut<RulesetCreated> for RulesetCreated {
     fn as_mut(&mut self) -> &mut RulesetCreated {
         self
@@ -856,6 +1001,7 @@ fn ruleset_unsupported() {
         }
     );
 
+    // Missing handled access because of the compatibility level.
     matches!(
         Ruleset::from(ABI::Unsupported)
             // HardRequirement for Ruleset.
@@ -865,6 +1011,16 @@ fn ruleset_unsupported() {
         RulesetError::CreateRuleset(CreateRulesetError::MissingHandledAccess)
     );
 
+    // Missing scope access because of the compatibility level.
+    matches!(
+        Ruleset::from(ABI::Unsupported)
+            // HardRequirement for Ruleset.
+            .set_compatibility(CompatLevel::HardRequirement)
+            .scope(Scope::Signal)
+            .unwrap_err(),
+        RulesetError::CreateRuleset(CreateRulesetError::MissingHandledAccess)
+    );
+
     assert_eq!(
         Ruleset::from(ABI::Unsupported)
             .handle_access(AccessFs::Execute)
@@ -920,6 +1076,7 @@ fn ruleset_unsupported() {
         }
     );
 
+    // Checks empty handled access with moot ruleset.
     assert!(matches!(
         Ruleset::from(ABI::Unsupported)
             // Empty access-rights
@@ -932,12 +1089,13 @@ fn ruleset_unsupported() {
 
     assert!(matches!(
         Ruleset::from(ABI::Unsupported)
-            // No handle_access() call.
+            // No handle_access() nor scope() call.
             .create()
             .unwrap_err(),
         RulesetError::CreateRuleset(CreateRulesetError::MissingHandledAccess)
     ));
 
+    // Checks empty handled access with minimal ruleset.
     assert!(matches!(
         Ruleset::from(ABI::V1)
             // Empty access-rights
@@ -948,6 +1106,22 @@ fn ruleset_unsupported() {
         )))
     ));
 
+    // Checks empty scope with moot ruleset.
+    assert!(matches!(
+        Ruleset::from(ABI::Unsupported)
+            .scope(Scope::from_all(ABI::Unsupported))
+            .unwrap_err(),
+        RulesetError::Scope(ScopeError::Compat(CompatError::Access(AccessError::Empty)))
+    ));
+
+    // Checks empty scope with minimal ruleset.
+    assert!(matches!(
+        Ruleset::from(ABI::V1)
+            .scope(Scope::from_all(ABI::Unsupported))
+            .unwrap_err(),
+        RulesetError::Scope(ScopeError::Compat(CompatError::Access(AccessError::Empty)))
+    ));
+
     // Tests inconsistency between the ruleset handled access-rights and the rule access-rights.
     for handled_access in &[
         make_bitflags!(AccessFs::{Execute | WriteFile}),
@@ -958,7 +1132,7 @@ fn ruleset_unsupported() {
             .unwrap();
         // Fakes a call to create() to test without involving the kernel (i.e. no
         // landlock_ruleset_create() call).
-        let ruleset_created = RulesetCreated::new(ruleset, -1);
+        let ruleset_created = RulesetCreated::new(ruleset, None);
         assert!(matches!(
             ruleset_created
                 .add_rule(PathBeneath::new(
@@ -1016,3 +1190,15 @@ fn unsupported_handled_access() {
         )))
     );
 }
+
+#[test]
+fn unsupported_handled_access_errno() {
+    assert_eq!(
+        Errno::from(
+            Ruleset::from(ABI::V3)
+                .handle_access(AccessNet::from_all(ABI::V3))
+                .unwrap_err()
+        ),
+        Errno::new(libc::EINVAL)
+    );
+}
